@@ -1,4 +1,4 @@
-# Copyright © 2018-2019 Roel van der Goot
+# Copyright © 2018-2020 Roel van der Goot
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -38,10 +38,20 @@ from ajsonapi.relationships import (
     OneToOneLocalRelationship,
     OneToOneRemoteRelationship,
 )
-from ajsonapi.types import Float32, Float64, Int32, Int64, String
+from ajsonapi.types import (
+    Boolean,
+    DateTime,
+    DateTimeTimeZone,
+    Float32,
+    Float64,
+    Int32,
+    Int64,
+    Json,
+    String,
+)
 
 assert version_info >= (3, 7), 'Need Python 3.7+ for module ajsonapi'
 del version_info
 
 __docformat__ = 'restructuredtext'
-__version__ = '0.4.0'
+__version__ = '0.6.3'
