@@ -1,4 +1,4 @@
-# Copyright © 2018-2019 Roel van der Goot
+# Copyright © 2018-2020 Roel van der Goot
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -22,7 +22,7 @@
 
 from asyncio import sleep
 
-from aiohttp import web
+from aiohttp import WSMsgType, web
 from asyncpg.exceptions import (
     CannotConnectNowError,
     ConnectionDoesNotExistError,
@@ -42,6 +42,7 @@ from ajsonapi.requests import (
 )
 from ajsonapi.responses import errors_response
 from ajsonapi.table import create_all as table_create_all
+from ajsonapi.table import parse_model
 from ajsonapi.uri.relationship import parse as parse_relationship
 
 
@@ -53,6 +54,7 @@ class Application:
 
     def __init__(self, *args, **kwargs):
         self.app = web.Application(*args, **kwargs)
+        self.ws_list = []
         self.pool = None
 
     async def connect_database(self, database):
@@ -81,8 +83,14 @@ class Application:
         """Disconnects from the database."""
         await self.pool.close()
 
+    def verify_model(self):
+        """Parses the model."""
+        # pylint: disable=no-self-use
+        parse_model()
+
     async def create_tables(self):
         """Creates all tables in the database."""
+        # pylint: disable=no-self-use
         await table_create_all()
 
     def add_json_api_routes(self):
@@ -117,6 +125,12 @@ class Application:
             web.get(path, self.get_related_resource),
         ])
 
+    def add_event_socket(self):
+        """Add event socket."""
+
+        path = '/event-socket'
+        self.app.add_routes([web.get(path, self.get_event_socket)])
+
     async def get_collection(self, request):
         """Entry point for GET /{collection} requests.
 
@@ -129,7 +143,8 @@ class Application:
                                                        allow_include=True,
                                                        allow_fields=True,
                                                        allow_filter=True,
-                                                       allow_sort=True)
+                                                       allow_sort=True,
+                                                       allow_page=True)
             return await collection.get(query)
         except ErrorsException as err:
             return errors_response(err.errors)
@@ -142,9 +157,9 @@ class Application:
         """
 
         try:
-            collection, data, query = await parse_collection_document_query(
+            collection, document, query = await parse_collection_document_query(
                 request)
-            return await collection.post(data, query)
+            return await collection.post(document, query)
         except ErrorsException as err:
             return errors_response(err.errors)
 
@@ -181,8 +196,9 @@ class Application:
             request: The DELETE /{collection}/{id} request.
         """
         try:
-            object_, data, query = await parse_object_document_query(request)
-            return await object_.patch(data, query)
+            object_, document, query = await parse_object_document_query(request
+                                                                        )
+            return await object_.patch(document, query)
         except ErrorsException as err:
             return errors_response(err.errors)
 
@@ -216,9 +232,9 @@ class Application:
             Appropriate response for the request.
         """
         try:
-            relationship, data, query = await parse_relationship_document_query(
-                request)
-            return await relationship.patch(data, query)
+            relationship, document, query = \
+                    await parse_relationship_document_query(request)
+            return await relationship.patch(document, query)
         except ErrorsException as err:
             return errors_response(err.errors)
 
@@ -234,9 +250,9 @@ class Application:
             Appropriate response for the request.
         """
         try:
-            relationship, data, query = await parse_relationship_document_query(
-                request)
-            return await relationship.post(data, query)
+            relationship, document, query = \
+                    await parse_relationship_document_query(request)
+            return await relationship.post(document, query)
         except ErrorsException as err:
             return errors_response(err.errors)
 
@@ -252,9 +268,9 @@ class Application:
             Appropriate response for the request.
         """
         try:
-            relationship, data, query = await parse_relationship_document_query(
-                request)
-            return await relationship.delete(data, query)
+            relationship, document, query = \
+                    await parse_relationship_document_query(request)
+            return await relationship.delete(document, query)
         except ErrorsException as err:
             return errors_response(err.errors)
 
@@ -287,7 +303,27 @@ class Application:
                 allow_include=True,
                 allow_fields=True,
                 allow_filter=True,
-                allow_sort=True)
+                allow_sort=True,
+                allow_page=True)
             return await related.get(query)
         except ErrorsException as err:
             return errors_response(err.errors)
+
+    async def get_event_socket(self, request):
+        """Creates a web socket for events."""
+
+        ws = web.WebSocketResponse()  # pylint: disable=invalid-name
+        await ws.prepare(request)
+        self.ws_list.append(ws)
+        async for msg in ws:
+            if msg.type == WSMsgType.TEXT:
+                if msg.data == 'close':
+                    await ws.close()
+                # Eventually: allow clients to register/deregister individual
+                # events they are interested in.
+            elif msg.type == WSMsgType.ERROR:
+                print('websocket connection closed with exception',
+                      ws.exception())
+        print('websocket connection closed')
+        self.ws_list.remove(ws)
+        return ws
