@@ -1,4 +1,4 @@
-# Copyright © 2018-2019 Roel van der Goot
+# Copyright © 2018-2020 Roel van der Goot
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -126,12 +126,12 @@ class IncludeInvalidPathError(Error):
     def to_error(self):
         return {
             'status':
-            '400',
+                '400',
             'source': {
                 'parameter': 'include'
             },
             'title':
-            'Invalid query parameter.',
+                'Invalid query parameter.',
             'detail': ("The resource does not have a "
                        f"'{self.path}' relationship path.")
         }
@@ -172,12 +172,12 @@ class FieldsInvalidFieldError(Error):
     def to_error(self):
         return {
             'status':
-            '400',
+                '400',
             'source': {
                 'parameter': 'fields'
             },
             'title':
-            'Invalid query parameter.',
+                'Invalid query parameter.',
             'detail': (f"Field '{self.field}' does not exist in resource "
                        f"'{self.resource}'.")
         }
@@ -449,16 +449,13 @@ class DocumentDataRelationshipConflictError(DocumentError):
 
     def to_error(self):
         return {
-            'status':
-            '409',
+            'status': '409',
             'source': {
                 'pointer': self.pointer,
             },
-            'title':
-            'Conflicting resource identifier object.',
-            'detail':
-            'Resource identifier object is part of another '
-            'one-to-one relationship.',
+            'title': 'Conflicting resource identifier object.',
+            'detail': 'Resource identifier object is part of another '
+                      'one-to-one relationship.',
         }
 
 
