@@ -1,4 +1,4 @@
-# Copyright © 2018-2019 Roel van der Goot
+# Copyright © 2018-2020 Roel van der Goot
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -36,6 +36,7 @@ from ajsonapi.errors import (
 from ajsonapi.exceptions import NonexistentRelationshipIdError
 from ajsonapi.field import Field
 from ajsonapi.id import Id
+from ajsonapi.logging import log
 from ajsonapi.uri.related import (
     ManyToManyRelatedResource,
     ManyToOneRelatedResource,
@@ -95,13 +96,13 @@ class Relationship(Field):
         """
 
     @abstractmethod
-    async def collection_remote_ids(self, local_ids=None):
+    def collection_remote_ids(self, query, local_ids=None):
         """Returns the SQL statement for the remote ids of a relationship's
         local collection.
         """
 
     @abstractmethod
-    async def object_remote_ids(self, local_ids=None):
+    def object_remote_ids(self, local_ids=None):
         """Returns the SQL condition for an ?include parameter from an
         object.
         """
@@ -191,8 +192,9 @@ class ToOneRelationship(Relationship):
 
     async def verify_data_ids_exist(self, env, data_ids):
         data_id_number = data_ids.number
-        stmt = f'SELECT FROM {self.rtable.name} WHERE id=$1;'
+        stmt = f'SELECT FROM {self.rtable.name} WHERE id=$1'
         async with env.lock:
+            log.debug(f'{stmt}: {data_id_number}')
             records = await env.connection.fetch(stmt, data_id_number)
         if records == []:
             data_id_name = data_ids.name
@@ -260,8 +262,9 @@ class ToManyRelationship(Relationship):
         data_id_numbers = [data_id.number for data_id in data_ids]
         stmt = (f'SELECT id '
                 f'FROM {self.rtable.name} '
-                f'WHERE id=ANY($1::UUID[]);')
+                f'WHERE id=ANY($1::UUID[])')
         async with env.lock:
+            log.debug(f'{stmt}: {data_id_numbers}')
             records = await env.connection.fetch(stmt, data_id_numbers)
         existing_id_names = {
             id_number_to_name(record['id']) for record in records
@@ -286,9 +289,7 @@ class OneToOneLocalRelationship(LocalRelationship, ToOneRelationship):
 
     def is_reverse(self, other):
         return (isinstance(other, OneToOneRemoteRelationship) and
-                other.table and self.rtable and
-                other.table.name == self.rtable.name and other.rtable and
-                self.table and other.rtable.name == self.table.name and
+                other.table is self.rtable and other.rtable is self.table and
                 other.rfkey == self.lfkey)
 
     def sql_constraints(self):
@@ -299,11 +300,32 @@ class OneToOneLocalRelationship(LocalRelationship, ToOneRelationship):
         return (f'UNIQUE ({self.lfkey}),\n'
                 f'FOREIGN KEY ({self.lfkey}) REFERENCES {self.rtable.name}(id)')
 
-    def collection_remote_ids(self, local_ids=None):
-        if local_ids is None:
+    def collection_remote_ids(self, query, local_ids=None):
+        filters = query['filters']
+        page = query['page']
+        if local_ids is None and filters is None and page == {}:
             partial = ''
         else:
-            partial = f' WHERE id IN ({local_ids})'
+            sort = query['sort']
+            if filters:
+                if sort:
+                    stmt = (f"SELECT id FROM {self.table.name} "
+                            f"WHERE {' AND '.join(filters)} "
+                            f"ORDER BY {', '.join(sort)}")
+                else:
+                    stmt = (f"SELECT id FROM {self.table.name} "
+                            f"WHERE {' AND '.join(filters)}")
+            else:
+                if sort:
+                    stmt = (f"SELECT id FROM {self.table.name} "
+                            f"ORDER BY {', '.join(sort)}")
+                else:
+                    stmt = f"SELECT id FROM {self.table.name}"
+            if page:
+                limit = page['size']
+                offset = page['number'] * limit
+                stmt += f' LIMIT {limit} OFFSET {offset}'
+            partial = f' WHERE id IN ({stmt})'
         return f'SELECT {self.lfkey} FROM {self.table.name}{partial}'
 
     def object_remote_ids(self, local_ids=None):
@@ -335,17 +357,36 @@ class OneToOneRemoteRelationship(ToOneRelationship):
         return OneToOneRemoteRelatedResource(name, collection, id_, self)
 
     def is_reverse(self, other):
-        return (isinstance(other, OneToOneLocalRelationship) and other.table and
-                self.rtable and other.table.name == self.rtable.name and
-                other.rtable and self.table and
-                other.rtable.name == self.table.name and
+        return (isinstance(other, OneToOneLocalRelationship) and
+                other.table is self.rtable and other.rtable is self.table and
                 other.lfkey == self.rfkey)
 
-    def collection_remote_ids(self, local_ids=None):
-        if local_ids is None:
+    def collection_remote_ids(self, query, local_ids=None):
+        filters = query['filters']
+        page = query['page']
+        if local_ids is None and filters is None and page == {}:
             partial = ' IS NOT NULL'
         else:
-            partial = f' IN ({local_ids})'
+            sort = query['sort']
+            if filters:
+                if sort:
+                    stmt = (f"SELECT id FROM {self.table.name} "
+                            f"WHERE {' AND '.join(filters)} "
+                            f"ORDER BY {', '.join(sort)}")
+                else:
+                    stmt = (f"SELECT id FROM {self.table.name} "
+                            f"WHERE {' AND '.join(filters)}")
+            else:
+                if sort:
+                    stmt = (f"SELECT id FROM {self.table.name} "
+                            f"ORDER BY {', '.join(sort)}")
+                else:
+                    stmt = f"SELECT id FROM {self.table.name}"
+            if page:
+                limit = page['size']
+                offset = page['number'] * limit
+                stmt += f' LIMIT {limit} OFFSET {offset}'
+            partial = f' IN ({stmt})'
         return (f'SELECT id FROM {self.rtable.name} '
                 f'WHERE {self.rfkey}{partial}')
 
@@ -386,25 +427,27 @@ class OneToOneRemoteRelationship(ToOneRelationship):
         """
         stmt = (f'SELECT {self.rfkey}, id '
                 f'FROM {self.rtable.name} '
-                f'WHERE {self.rfkey} IS NOT NULL;')
+                f'WHERE {self.rfkey} IS NOT NULL')
         async with env.lock:
+            log.debug(stmt)
             records = await env.connection.fetch(stmt)
         await env.event.wait()
         rid_by_lid = {record[self.rfkey]: record['id'] for record in records}
         for lid, data in env.data_by_lid.items():
             if lid in rid_by_lid:
                 relationship_data = {
-                    'type': self.reverse.collection.name,
+                    'type': self.rtable.collection.name,
                     'id': id_number_to_name(rid_by_lid[lid]),
                 }
             else:
                 relationship_data = None
+            lid_name = id_number_to_name(lid)
             data.setdefault('relationships', {})[self.name] = {
                 'data': relationship_data,
                 'links': {
-                    'self': (f'/{self.collection.name}/{lid}/relationships/'
-                             f'{self.name}'),
-                    'related': (f'/{self.collection.name}/{lid}/'
+                    'self': (f'/{self.collection.name}/{lid_name}/'
+                             f'relationships/{self.name}'),
+                    'related': (f'/{self.collection.name}/{lid_name}/'
                                 f'{self.name}'),
                 },
             }
@@ -415,25 +458,27 @@ class OneToOneRemoteRelationship(ToOneRelationship):
         """
         stmt = (f'SELECT {self.rfkey}, id '
                 f'FROM {self.rtable.name} '
-                f'WHERE {self.rfkey} IN ({ids});')
+                f'WHERE {self.rfkey} IN ({ids})')
         async with env.lock:
+            log.debug(stmt)
             records = await env.connection.fetch(stmt)
         await env.event.wait()
         rid_by_lid = {record[self.rfkey]: record['id'] for record in records}
         for lid, data in env.included_by_id.items():
             if lid in rid_by_lid:
                 relationship_data = {
-                    'type': self.reverse.collection.name,
+                    'type': self.rtable.collection.name,
                     'id': id_number_to_name(rid_by_lid[lid])
                 }
             else:
                 relationship_data = None
+            lid_name = id_number_to_name(lid)
             data.setdefault('relationships', {})[self.name] = {
                 'data': relationship_data,
                 'links': {
-                    'self': (f'/{self.collection.name}/{lid}/relationships/'
-                             f'{self.name}'),
-                    'related': (f'/{self.collection.name}/{lid}/'
+                    'self': (f'/{self.collection.name}/{lid_name}/'
+                             f'relationships/{self.name}'),
+                    'related': (f'/{self.collection.name}/{lid_name}/'
                                 f'{self.name}'),
                 },
             }
@@ -444,11 +489,12 @@ class OneToOneRemoteRelationship(ToOneRelationship):
         """
         stmt = (f'UPDATE {self.rtable.name} '
                 f'SET {self.rfkey} = $1 '
-                f'WHERE id=$2;')
+                f'WHERE id=$2')
         data_id = env.relationships[self.name]
         data_id_number = data_id.number
         await env.event.wait()
         async with env.lock:
+            log.debug(f'{stmt}: {env.id_number}, {data_id_number}')
             result = await env.connection.execute(stmt, env.id_number,
                                                   data_id_number)
         if result == 'UPDATE 0':
@@ -478,18 +524,19 @@ class OneToOneRemoteRelationship(ToOneRelationship):
         """
         stmt = (f"SELECT id "
                 f"FROM {self.rtable.name} "
-                f"WHERE {self.rfkey} = $1;")
+                f"WHERE {self.rfkey} = $1")
         async with env.lock:
+            log.debug(f'{stmt}: {env.obj.id.number}')
             record = await env.connection.fetchrow(stmt, env.obj.id.number)
         await env.event.wait()
         if record:
             rel_data = {
-                'type': self.reverse.collection.name,
-                'id': record['id'],
+                'type': self.rtable.collection.name,
+                'id': id_number_to_name(record['id']),
             }
         else:
             rel_data = None
-        env.data['relationships'][self.name] = {
+        env.data.setdefault('relationships', {})[self.name] = {
             'data': rel_data,
             'links': {
                 'self': (f'/{self.collection.name}/{env.obj.id.name}/'
@@ -505,25 +552,27 @@ class OneToOneRemoteRelationship(ToOneRelationship):
         """
         stmt = (f'SELECT {self.rfkey}, id '
                 f'FROM {self.rtable.name} '
-                f'WHERE {self.rfkey} IN ({ids});')
+                f'WHERE {self.rfkey} IN ({ids})')
         async with env.lock:
+            log.debug(f'{stmt}: {env.object_id}')
             records = await env.connection.fetch(stmt, env.object_id)
         await env.event.wait()
         rid_by_lid = {record[self.rfkey]: record['id'] for record in records}
         for lid, data in env.included_by_id.items():
             if lid in rid_by_lid:
                 relationship_data = {
-                    'type': self.reverse.collection.name,
+                    'type': self.rtable.collection.name,
                     'id': id_number_to_name(rid_by_lid[lid])
                 }
             else:
                 relationship_data = None
+            lid_name = id_number_to_name(lid)
             data.setdefault('relationships', {})[self.name] = {
                 'data': relationship_data,
                 'links': {
-                    'self': (f'/{self.collection.name}/{lid}/relationships/'
-                             f'{self.name}'),
-                    'related': (f'/{self.collection.name}/{lid}/'
+                    'self': (f'/{self.collection.name}/{lid_name}/'
+                             f'relationships/{self.name}'),
+                    'related': (f'/{self.collection.name}/{lid_name}/'
                                 f'{self.name}'),
                 }
             }
@@ -534,10 +583,11 @@ class OneToOneRemoteRelationship(ToOneRelationship):
         """
         stmt = (f'UPDATE {self.rtable.name} '
                 f'SET {self.rfkey} = $1 '
-                f'WHERE id=$2;')
+                f'WHERE id=$2')
         data_id = env.relationships[self.name]
         data_id_number = data_id.number
         async with env.lock:
+            log.debug(f'{stmt}: {env.obj.id.number}, {data_id_number}')
             result = await env.connection.execute(stmt, env.obj.id.number,
                                                   data_id_number)
         if result == 'UPDATE 0':
@@ -571,7 +621,7 @@ class OneToOneRemoteRelationship(ToOneRelationship):
         }
         """
         if isinstance(env.related.relationship, LocalRelationship):
-            # OneToOneLocalRelationship or ManyToOneLocalRelationship
+            # OneToOneLocalRelationship or ManyToOneRelationship
             stmt = (f'SELECT id '
                     f'FROM {self.rtable.name} '
                     f'WHERE {self.rfkey} IN ('
@@ -586,13 +636,14 @@ class OneToOneRemoteRelationship(ToOneRelationship):
                     f' FROM {env.related.relationship.rtable.name}'
                     f' WHERE {env.related.relationship.rfkey} = $1)')
         async with env.lock:
+            log.debug(f'{stmt}: {env.related.id.number}')
             record = await env.connection.fetchrow(stmt, env.related.id.number)
         await env.event.wait()
         if env.data is None:
             return
         if record:
             rel_data = {
-                'type': self.reverse.collection.name,
+                'type': self.rtable.collection.name,
                 'id': id_number_to_name(record[0]),
             }
         else:
@@ -646,6 +697,7 @@ class OneToOneRemoteRelationship(ToOneRelationship):
                     f' FROM {env.related.relationship.rtable.name}'
                     f' WHERE {env.related.relationship.rfkey} = $1)')
         async with env.lock:
+            log.debug(f'{stmt}: {env.related.id.number}')
             records = await env.connection.fetch(stmt, env.related.id.number)
         await env.event.wait()
         rids_by_lid = {}
@@ -656,7 +708,7 @@ class OneToOneRemoteRelationship(ToOneRelationship):
                 rids = rids_by_lid[lid]
                 assert len(rids) == 1
                 relationship_data = {
-                    'type': self.reverse.collection.name,
+                    'type': self.rtable.collection.name,
                     'id': id_number_to_name(rids[0]),
                 }
             else:
@@ -690,17 +742,36 @@ class OneToManyRelationship(ToManyRelationship):
         return OneToManyRelatedResource(name, collection, id_, self)
 
     def is_reverse(self, other):
-        return (isinstance(other, ManyToOneRelationship) and other.table and
-                self.rtable and other.table.name == self.rtable.name and
-                other.rtable and self.table and
-                other.rtable.name == self.table.name and
+        return (isinstance(other, ManyToOneRelationship) and
+                other.table is self.rtable and other.rtable is self.table and
                 other.lfkey == self.rfkey)
 
-    def collection_remote_ids(self, local_ids=None):
-        if local_ids is None:
+    def collection_remote_ids(self, query, local_ids=None):
+        filters = query['filters']
+        page = query['page']
+        if local_ids is None and filters is None and page == {}:
             partial = ' IS NOT NULL'
         else:
-            partial = f' IN ({local_ids})'
+            sort = query['sort']
+            if filters:
+                if sort:
+                    stmt = (f"SELECT id FROM {self.table.name} "
+                            f"WHERE {' AND '.join(filters)} "
+                            f"ORDER BY {', '.join(sort)}")
+                else:
+                    stmt = (f"SELECT id FROM {self.table.name} "
+                            f"WHERE {' AND '.join(filters)}")
+            else:
+                if sort:
+                    stmt = (f"SELECT id FROM {self.table.name} "
+                            f"ORDER BY {', '.join(sort)}")
+                else:
+                    stmt = f"SELECT id FROM {self.table.name}"
+            if page:
+                limit = page['size']
+                offset = page['number'] * limit
+                stmt += f' LIMIT {limit} OFFSET {offset}'
+            partial = f' IN ({stmt})'
         return (f'SELECT id FROM {self.rtable.name} '
                 f'WHERE {self.rfkey}{partial}')
 
@@ -746,6 +817,7 @@ class OneToManyRelationship(ToManyRelationship):
                 f'FROM {self.rtable.name} '
                 f'WHERE {self.rfkey} IS NOT NULL')
         async with env.lock:
+            log.debug(stmt)
             records = await env.connection.fetch(stmt)
         await env.event.wait()
         rids_by_lid = {}
@@ -754,17 +826,18 @@ class OneToManyRelationship(ToManyRelationship):
         for lid, data in env.data_by_lid.items():
             if lid in rids_by_lid:
                 relationship_data = [{
-                    'type': self.reverse.collection.name,
+                    'type': self.rtable.collection.name,
                     'id': id_number_to_name(rid),
                 } for rid in rids_by_lid[lid]]
             else:
                 relationship_data = []
+            lid_name = id_number_to_name(lid)
             data.setdefault('relationships', {})[self.name] = {
                 'data': relationship_data,
                 'links': {
-                    'self': (f'/{self.collection.name}/{lid}/relationships/'
-                             f'{self.name}'),
-                    'related': (f'/{self.collection.name}/{lid}/'
+                    'self': (f'/{self.collection.name}/{lid_name}/'
+                             f'relationships/{self.name}'),
+                    'related': (f'/{self.collection.name}/{lid_name}/'
                                 f'{self.name}'),
                 }
             }
@@ -775,8 +848,9 @@ class OneToManyRelationship(ToManyRelationship):
         """
         stmt = (f'SELECT {self.rfkey}, id '
                 f'FROM {self.rtable.name} '
-                f'WHERE {self.rfkey} IN ({ids});')
+                f'WHERE {self.rfkey} IN ({ids})')
         async with env.lock:
+            log.debug(stmt)
             records = await env.connection.fetch(stmt)
         await env.event.wait()
         rids_by_lid = {}
@@ -785,17 +859,18 @@ class OneToManyRelationship(ToManyRelationship):
         for lid, data in env.included_by_id.items():
             if lid in rids_by_lid:
                 relationship_data = [{
-                    'type': self.reverse.collection.name,
+                    'type': self.rtable.collection.name,
                     'id': id_number_to_name(rid),
                 } for rid in rids_by_lid[lid]]
             else:
                 relationship_data = []
+            lid_name = id_number_to_name(lid)
             data.setdefault('relationships', {})[self.name] = {
                 'data': relationship_data,
                 'links': {
-                    'self': (f'/{self.collection.name}/{lid}/relationships/'
-                             f'{self.name}'),
-                    'related': (f'/{self.collection.name}/{lid}/'
+                    'self': (f'/{self.collection.name}/{lid_name}/'
+                             f'relationships/{self.name}'),
+                    'related': (f'/{self.collection.name}/{lid_name}/'
                                 f'{self.name}'),
                 }
             }
@@ -806,11 +881,12 @@ class OneToManyRelationship(ToManyRelationship):
         """
         stmt = (f'UPDATE {self.rtable.name} '
                 f'SET {self.rfkey} = $1 '
-                f'WHERE id=ANY($2::UUID[]);')
+                f'WHERE id=ANY($2::UUID[])')
         data_ids = env.relationships[self.name]
         data_id_numbers = [data_id.number for data_id in data_ids]
         await env.event.wait()
         async with env.lock:
+            log.debug(f'{stmt}: {env.id_number}, {data_id_numbers}')
             result = await env.connection.execute(stmt, env.id_number,
                                                   data_id_numbers)
         if result == 'UPDATE 0':
@@ -844,14 +920,15 @@ class OneToManyRelationship(ToManyRelationship):
         lfkey_name = self.rfkey
         stmt = (f"SELECT id "
                 f"FROM {self.rtable.name} "
-                f"WHERE {lfkey_name} = $1;")
+                f"WHERE {lfkey_name} = $1")
         async with env.lock:
+            log.debug(f'{stmt}: {env.obj.id.number}')
             records = await env.connection.fetch(stmt, env.obj.id.number)
         await env.event.wait()
         rids = [record['id'] for record in records]
-        env.data['relationships'][self.name] = {
+        env.data.setdefault('relationships', {})[self.name] = {
             'data': [{
-                'type': self.reverse.collection.name,
+                'type': self.rtable.collection.name,
                 'id': id_number_to_name(rid),
             } for rid in rids],
             'links': {
@@ -868,8 +945,9 @@ class OneToManyRelationship(ToManyRelationship):
         """
         stmt = (f'SELECT {self.rfkey}, id '
                 f'FROM {self.rtable.name} '
-                f'WHERE {self.rfkey} IN ({ids});')
+                f'WHERE {self.rfkey} IN ({ids})')
         async with env.lock:
+            log.debug(f'{stmt}: {env.object_id}')
             records = await env.connection.fetch(stmt, env.object_id)
         await env.event.wait()
         rids_by_lid = {}
@@ -878,17 +956,18 @@ class OneToManyRelationship(ToManyRelationship):
         for lid, data in env.included_by_id.items():
             if lid in rids_by_lid:
                 relationship_data = [{
-                    'type': self.reverse.collection.name,
+                    'type': self.rtable.collection.name,
                     'id': id_number_to_name(rid),
                 } for rid in rids_by_lid[lid]]
             else:
                 relationship_data = []
+            lid_name = id_number_to_name(lid)
             data.setdefault('relationships', {})[self.name] = {
                 'data': relationship_data,
                 'links': {
-                    'self': (f'/{self.collection.name}/{lid}/relationships/'
-                             f'{self.name}'),
-                    'related': (f'/{self.collection.name}/{lid}/'
+                    'self': (f'/{self.collection.name}/{lid_name}/'
+                             f'relationships/{self.name}'),
+                    'related': (f'/{self.collection.name}/{lid_name}/'
                                 f'{self.name}'),
                 }
             }
@@ -899,10 +978,11 @@ class OneToManyRelationship(ToManyRelationship):
         """
         stmt = (f'UPDATE {self.rtable.name} '
                 f'SET {self.rfkey} = $1 '
-                f'WHERE id=ANY($2::UUID[]);')
+                f'WHERE id=ANY($2::UUID[])')
         data_ids = env.relationships[self.name]
         data_id_numbers = [data_id.number for data_id in data_ids]
         async with env.lock:
+            log.debug(f'{stmt}: {env.obj.id.number}, {data_id_numbers}')
             result = await env.connection.execute(stmt, env.obj.id.number,
                                                   data_id_numbers)
         if result == 'UPDATE 0':
@@ -953,13 +1033,14 @@ class OneToManyRelationship(ToManyRelationship):
                     f' FROM {env.related.relationship.rtable.name}'
                     f' WHERE {env.related.relationship.rfkey} = $1)')
         async with env.lock:
+            log.debug(f'{stmt}: {env.related.id.number}')
             records = await env.connection.fetch(stmt, env.related.id.number)
         await env.event.wait()
         if env.data is None:
             return
         rel_data = [{
             'type': self.rtable.collection.name,
-            'id': record[0]
+            'id': id_number_to_name(record[0])
         } for record in records]
         env.data.setdefault('relationships', {})[self.name] = {
             'data': rel_data,
@@ -1011,6 +1092,7 @@ class OneToManyRelationship(ToManyRelationship):
                     f' FROM {env.related.relationship.rtable.name}'
                     f' WHERE {env.related.relationship.rfkey} = $1)')
         async with env.lock:
+            log.debug(f'{stmt}: {env.related.id.number}')
             records = await env.connection.fetch(stmt, env.related.id.number)
         await env.event.wait()
         rids_by_lid = {}
@@ -1019,17 +1101,18 @@ class OneToManyRelationship(ToManyRelationship):
         for lid, data in env.data_by_lid.items():
             if lid in rids_by_lid:
                 relationship_data = [{
-                    'type': self.reverse.collection.name,
+                    'type': self.rtable.collection.name,
                     'id': id_number_to_name(rid),
                 } for rid in rids_by_lid[lid]]
             else:
                 relationship_data = []
+            lid_name = id_number_to_name(lid)
             data.setdefault('relationships', {})[self.name] = {
                 'data': relationship_data,
                 'links': {
-                    'self': (f'/{self.collection.name}/{lid}/'
+                    'self': (f'/{self.collection.name}/{lid_name}/'
                              f'relationships/{self.name}'),
-                    'related': (f'/{self.collection.name}/{lid}/'
+                    'related': (f'/{self.collection.name}/{lid_name}/'
                                 f'{self.name}'),
                 },
             }
@@ -1045,10 +1128,8 @@ class ManyToOneRelationship(LocalRelationship, ToOneRelationship):
         return ManyToOneRelatedResource(name, collection, id_, self)
 
     def is_reverse(self, other):
-        return (isinstance(other, OneToManyRelationship) and other.table and
-                self.rtable and other.table.name == self.rtable.name and
-                not isinstance(other.rtable, str) and self.table and
-                other.rtable.name == self.table.name and
+        return (isinstance(other, OneToManyRelationship) and
+                other.rtable is self.table and other.rfkey == self.lfkey and
                 other.rfkey == self.lfkey)
 
     def sql_constraints(self):
@@ -1058,11 +1139,32 @@ class ManyToOneRelationship(LocalRelationship, ToOneRelationship):
 
         return f'FOREIGN KEY ({self.lfkey}) REFERENCES {self.rtable.name}(id)'
 
-    def collection_remote_ids(self, local_ids=None):
-        if local_ids is None:
+    def collection_remote_ids(self, query, local_ids=None):
+        filters = query['filters']
+        page = query['page']
+        if local_ids is None and filters is None and page == {}:
             partial = ''
         else:
-            partial = f' WHERE id IN ({local_ids})'
+            sort = query['sort']
+            if filters:
+                if sort:
+                    stmt = (f"SELECT id FROM {self.table.name} "
+                            f"WHERE {' AND '.join(filters)} "
+                            f"ORDER BY {', '.join(sort)}")
+                else:
+                    stmt = (f"SELECT id FROM {self.table.name} "
+                            f"WHERE {' AND '.join(filters)}")
+            else:
+                if sort:
+                    stmt = (f"SELECT id FROM {self.table.name} "
+                            f"ORDER BY {', '.join(sort)}")
+                else:
+                    stmt = f"SELECT id FROM {self.table.name}"
+            if page:
+                limit = page['size']
+                offset = page['number'] * limit
+                stmt += f' LIMIT {limit} OFFSET {offset}'
+            partial = f' WHERE id IN ({stmt})'
         return f'SELECT {self.lfkey} FROM {self.table.name}{partial}'
 
     def object_remote_ids(self, local_ids=None):
@@ -1096,18 +1198,37 @@ class ManyToManyRelationship(ToManyRelationship):
         return ManyToManyRelatedResource(name, collection, id_, self)
 
     def is_reverse(self, other):
-        return (isinstance(other, ManyToManyRelationship) and other.table and
-                self.rtable and other.table.name == self.rtable.name and
-                other.rtable and self.table and
-                other.rtable.name == self.table.name and
-                other.atable == self.atable and other.lafkey == self.rafkey and
+        return (isinstance(other, ManyToManyRelationship) and
+                other.table is self.rtable and other.rtable is self.table and
+                other.atable is self.atable and other.lafkey == self.rafkey and
                 other.rafkey == self.lafkey)
 
-    def collection_remote_ids(self, local_ids=None):
-        if local_ids is None:
+    def collection_remote_ids(self, query, local_ids=None):
+        filters = query['filters']
+        page = query['page']
+        if local_ids is None and filters is None and page == {}:
             partial = ' IS NOT NULL'
         else:
-            partial = f' IN ({local_ids})'
+            sort = query['sort']
+            if filters:
+                if sort:
+                    stmt = (f"SELECT id FROM {self.table.name} "
+                            f"WHERE {' AND '.join(filters)} "
+                            f"ORDER BY {', '.join(sort)}")
+                else:
+                    stmt = (f"SELECT id FROM {self.table.name} "
+                            f"WHERE {' AND '.join(filters)}")
+            else:
+                if sort:
+                    stmt = (f"SELECT id FROM {self.table.name} "
+                            f"ORDER BY {', '.join(sort)}")
+                else:
+                    stmt = f"SELECT id FROM {self.table.name}"
+            if page:
+                limit = page['size']
+                offset = page['number'] * limit
+                stmt += f' LIMIT {limit} OFFSET {offset}'
+            partial = f' IN ({stmt})'
         return (f'SELECT {self.rafkey} FROM {self.atable.name} '
                 f'WHERE {self.lafkey}{partial}')
 
@@ -1150,8 +1271,9 @@ class ManyToManyRelationship(ToManyRelationship):
         }
         """
         stmt = (f'SELECT {self.lafkey}, {self.rafkey} '
-                f'FROM {self.atable.name};')
+                f'FROM {self.atable.name}')
         async with env.lock:
+            log.debug(stmt)
             records = await env.connection.fetch(stmt)
         await env.event.wait()
         rids_by_lid = {}
@@ -1161,17 +1283,18 @@ class ManyToManyRelationship(ToManyRelationship):
         for lid, data in env.data_by_lid.items():
             if lid in rids_by_lid:
                 relationship_data = [{
-                    'type': self.reverse.collection.name,
+                    'type': self.rtable.collection.name,
                     'id': id_number_to_name(rid),
                 } for rid in rids_by_lid[lid]]
             else:
                 relationship_data = []
+            lid_name = id_number_to_name(lid)
             data.setdefault('relationships', {})[self.name] = {
                 'data': relationship_data,
                 'links': {
-                    'self': (f'/{self.collection.name}/{lid}/'
+                    'self': (f'/{self.collection.name}/{lid_name}/'
                              f'relationships/{self.name}'),
-                    'related': (f'/{self.collection.name}/{lid}/'
+                    'related': (f'/{self.collection.name}/{lid_name}/'
                                 f'{self.name}'),
                 },
             }
@@ -1182,8 +1305,9 @@ class ManyToManyRelationship(ToManyRelationship):
         """
         stmt = (f'SELECT {self.lafkey}, {self.rafkey} '
                 f'FROM {self.atable.name} '
-                f'WHERE {self.lafkey} IN ({ids});')
+                f'WHERE {self.lafkey} IN ({ids})')
         async with env.lock:
+            log.debug(stmt)
             records = await env.connection.fetch(stmt)
         await env.event.wait()
         rids_by_lid = {}
@@ -1193,17 +1317,18 @@ class ManyToManyRelationship(ToManyRelationship):
         for lid, data in env.included_by_id.items():
             if lid in rids_by_lid:
                 relationship_data = [{
-                    'type': self.reverse.collection.name,
+                    'type': self.rtable.collection.name,
                     'id': id_number_to_name(rid),
                 } for rid in rids_by_lid[lid]]
             else:
                 relationship_data = []
+            lid_name = id_number_to_name(lid)
             data.setdefault('relationships', {})[self.name] = {
                 'data': relationship_data,
                 'links': {
-                    'self': (f'/{self.collection.name}/{lid}/'
+                    'self': (f'/{self.collection.name}/{lid_name}/'
                              f'relationships/{self.name}'),
-                    'related': (f'/{self.collection.name}/{lid}/'
+                    'related': (f'/{self.collection.name}/{lid_name}/'
                                 f'{self.name}'),
                 },
             }
@@ -1214,13 +1339,14 @@ class ManyToManyRelationship(ToManyRelationship):
         """
         stmt = (f'INSERT INTO {self.atable.name} '
                 f'({self.lafkey}, {self.rafkey}) '
-                f'VALUES ($1, $2);')
+                f'VALUES ($1, $2)')
         data_ids = env.relationships[self.name]
         await env.event.wait()
         data = [(env.id_number, data_id.number) for data_id in data_ids]
         try:
             async with env.lock:
                 async with env.connection.transaction():
+                    log.debug(f'{stmt}: {data}')
                     await env.connection.executemany(stmt, data)
         except ForeignKeyViolationError:
             env.cancel()
@@ -1252,7 +1378,7 @@ class ManyToManyRelationship(ToManyRelationship):
         """
         stmt = (f"SELECT {self.rafkey} "
                 f"FROM {self.atable.name} "
-                f"WHERE {self.lafkey} = $1;")
+                f"WHERE {self.lafkey} = $1")
         async with env.lock:
             records = await env.connection.fetch(stmt, env.obj.id.number)
         await env.event.wait()
@@ -1261,9 +1387,9 @@ class ManyToManyRelationship(ToManyRelationship):
         for record in records:
             rafkeys_by_lafkey.setdefault(env.obj.id.number,
                                          []).append(record[self.rafkey])
-        env.data['relationships'][self.name] = {
+        env.data.setdefault('relationships', {})[self.name] = {
             'data': [{
-                'type': self.reverse.collection.name,
+                'type': self.rtable.collection.name,
                 'id': id_number_to_name(rid),
             } for rid in rids],
             'links': {
@@ -1280,7 +1406,7 @@ class ManyToManyRelationship(ToManyRelationship):
         """
         stmt = (f'SELECT {self.lafkey}, {self.rafkey} '
                 f'FROM {self.atable.name} '
-                f'WHERE {self.lafkey} IN ({ids});')
+                f'WHERE {self.lafkey} IN ({ids})')
         async with env.lock:
             records = await env.connection.fetch(stmt, env.object_id)
         await env.event.wait()
@@ -1291,17 +1417,18 @@ class ManyToManyRelationship(ToManyRelationship):
         for lid, data in env.included_by_id.items():
             if lid in rids_by_lid:
                 relationship_data = [{
-                    'type': self.reverse.collection.name,
+                    'type': self.rtable.collection.name,
                     'id': id_number_to_name(rid),
                 } for rid in rids_by_lid[lid]]
             else:
                 relationship_data = []
+            lid_name = id_number_to_name(lid)
             data.setdefault('relationships', {})[self.name] = {
                 'data': relationship_data,
                 'links': {
-                    'self': (f'/{self.collection.name}/{lid}/'
+                    'self': (f'/{self.collection.name}/{lid_name}/'
                              f'relationships/{self.name}'),
-                    'related': (f'/{self.collection.name}/{lid}/'
+                    'related': (f'/{self.collection.name}/{lid_name}/'
                                 f'{self.name}'),
                 },
             }
@@ -1312,12 +1439,13 @@ class ManyToManyRelationship(ToManyRelationship):
         """
         stmt = (f'INSERT INTO {self.atable.name} '
                 f'({self.lafkey}, {self.rafkey}) '
-                f'VALUES ($1, $2);')
+                f'VALUES ($1, $2)')
         data_ids = env.relationships[self.name]
         data = [(env.obj.id.number, data_id.number) for data_id in data_ids]
         try:
             async with env.lock:
                 async with env.connection.transaction():
+                    log.debug(f'{stmt}: {data}')
                     await env.connection.executemany(stmt, data)
         except ForeignKeyViolationError:
             env.cancel()
@@ -1367,13 +1495,14 @@ class ManyToManyRelationship(ToManyRelationship):
                     f' FROM {env.related.relationship.rtable.name}'
                     f' WHERE {env.related.relationship.rfkey} = $1)')
         async with env.lock:
+            log.debug(f'{stmt}, {env.related.id.number}')
             records = await env.connection.fetch(stmt, env.related.id.number)
         await env.event.wait()
         if env.data is None:
             return
         rids = [record[0] for record in records]
         relationship_data = [{
-            'type': self.reverse.collection.name,
+            'type': self.rtable.collection.name,
             'id': id_number_to_name(rid),
         } for rid in rids]
         env.data.setdefault('relationships', {})[self.name] = {
@@ -1426,6 +1555,7 @@ class ManyToManyRelationship(ToManyRelationship):
                     f' FROM {env.related.relationship.rtable.name}'
                     f' WHERE {env.related.relationship.rfkey} = $1)')
         async with env.lock:
+            log.debug(f'{stmt}: {env.related.id.number}')
             records = await env.connection.fetch(stmt, env.related.id.number)
         await env.event.wait()
         rids_by_lid = {}
@@ -1434,7 +1564,7 @@ class ManyToManyRelationship(ToManyRelationship):
         for lid, data in env.data_by_lid.items():
             if lid in rids_by_lid:
                 relationship_data = [{
-                    'type': self.reverse.collection.name,
+                    'type': self.rtable.collection.name,
                     'id': id_number_to_name(rid),
                 } for rid in rids_by_lid[lid]]
             else:
@@ -1443,6 +1573,6 @@ class ManyToManyRelationship(ToManyRelationship):
                 'data': relationship_data,
                 'links': {
                     'self':
-                    f'/{env.related.collection.name}/{env.related.id.name}'
+                        f'/{env.related.collection.name}/{env.related.id.name}'
                 },
             }
