@@ -1,4 +1,4 @@
-# Copyright © 2018-2019 Roel van der Goot
+# Copyright © 2018-2020 Roel van der Goot
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -22,36 +22,89 @@
 """
 
 
-class Float32:
+class Type:
+    """Abstract base class for supported database types."""
+
+    # pylint: disable=too-few-public-methods
+
+    @staticmethod
+    def sql(value):
+        """Converts value into sql value."""
+        return value
+
+
+class Boolean(Type):
+    """Type for booleans."""
+
+    # pylint: disable=too-few-public-methods
+
+    name = 'BOOLEAN'
+
+
+class DateTime(Type):
+    """Type for date times."""
+
+    # pylint: disable=too-few-public-methods
+
+    name = 'TIMESTAMP'
+
+
+class DateTimeTimeZone(Type):
+    """Type for date times with time zone information."""
+
+    # pylint: disable=too-few-public-methods
+
+    name = 'TIMESTAMP WITH TIME ZONE'
+
+
+class Float32(Type):
     """Type for 32 bit floating point numbers."""
+
     # pylint: disable=too-few-public-methods
 
     name = 'float'
 
 
-class Float64:
+class Float64(Type):
     """Type for 64 bit floating point numbers."""
+
     # pylint: disable=too-few-public-methods
 
     name = 'float8'
 
 
-class Int32:
+class Int32(Type):
     """Type for 32 bit signed integers."""
+
     # pylint: disable=too-few-public-methods
 
     name = 'INT'
 
 
-class Int64:
+class Int64(Type):
     """Type for 64 bit signed integers."""
+
     # pylint: disable=too-few-public-methods
 
     name = 'BIGINT'
 
 
-class String:
+class String(Type):
     """Type for strings."""
+
     # pylint: disable=too-few-public-methods
 
     name = 'TEXT'
+
+    @staticmethod
+    def sql(value):
+        sql_value = value.replace("'", "''")
+        return f"'{sql_value}'"
+
+
+class Json(Type):
+    """Type for json data."""
+
+    # pylint: disable=too-few-public-methods
+
+    name = 'JSONB'
